/* DIPOLY.H: Header file for polygon-filling code */

/* Describes a single point (used for a single vertex) */
struct vertex {
   int iX;   /* X coordinate */
   int iY;   /* Y coordinate */
};

/* Describes a series of points (used to store a list of vertices that
   describe a polygon; each vertex is assumed to connect to the two
   adjacent vertices, and the last vertex is assumed to connect to the
   first) */
struct vertexListHeader {
   int iLength;                /* # of points */
   struct vertex * pvtx;   /* pointer to list of points */
};

/* Describes the beginning and ending X coordinates of a single
   horizontal line */
struct hLine {
   int iXStart; /* X coordinate of leftmost pixel in line */
   int iXEnd;   /* X coordinate of rightmost pixel in line */
};

/* Describes a Length-long series of horizontal lines, all assumed to
   be on contiguous scan lines starting at YStart and proceeding
   downward (used to describe a scan-converted polygon to the
   low-level hardware-dependent drawing code) */
struct hLineList {
   int iLength;                /* # of horizontal lines */
   int iYStart;                /* Y coordinate of topmost line */
   struct hLine * phln;   /* pointer to list of horz lines */
};

hLineList * phllScannedPolygon( vertexListHeader * pvtl, int iXOffset, int iYOffset );